local g = 9.81;
local ppm = 50; -- pixels per meter.

local function Gnomario_World_AtEdge(self)

end

local function Gnomario_World_OnGround(self)
	return self.grounded;
end
local function Gnomario_World_AtWall(self)
	return (self.grounded and self.obstructed);
end
local function Gnomario_World_Jump(self)
	if self.grounded then
		self.forceV = 5.75;
		PlaySoundFile([[Interface\AddOns\Gnomario\Sounds\jump.ogg]]);
	end
end


local function Apply_Physics(self)
	self.AtEdge = Gnomario_World_AtEdge;
	self.OnGround = Gnomario_World_OnGround;
	self.AtWall = Gnomario_World_AtWall;
	self.Jump = Gnomario_World_Jump;
end

local function Gnomario_Physics_Add(self, object)
	table.insert(self.objects, object);
end

local function Gnomario_Physics_Step(self, duration)
	for n = 1, #self.objects do
		local obj = self.objects[n];
		local x,y = obj.x, obj.y;
		if not obj.static then
			obj.forceV = (obj.forceV or 0) - ( (obj.g or g) * duration );
			if obj.forceV ~= 0 then
				obj.grounded = false;
			end
			if (obj.forceH < 0 and obj.limitH < 0) or (obj.forceH > 0 and obj.limitH > 0) then obj.forceH = 0; end
			if (obj.forceV < 0 and obj.limitV < 0) or (obj.forceV > 0 and obj.limitV > 0) then obj.forceV = 0; end
			y = y + ( ( (0 and obj.grounded) or obj.forceV ) * duration * ppm);
			if y <= 0 or (obj.landed and y <= (obj.landed.y + obj.landed.height)) then y = math.max(0, y); obj.grounded = true; obj.forceV = 0; end
			if ( obj.grounded ) then
				x = x + ( (obj.forceH or 0) * duration * ppm )
				obj.offsetH = obj.forceH;
			else
				x = x + ( (obj.offsetH or 0) * duration * ppm )
			end
			-- bla bla bla collision test goes here
			local x1, y1 = obj.x + (obj.width/2), obj.y + (obj.height/2);
			local collided = false;
			for m = 1, #self.objects do
				local cobj = self.objects[m];
				if cobj ~= obj then
					local x2, y2 = cobj.x + (cobj.width/2), cobj.y + (cobj.height/2);
					local dx, dy = math.abs(x2-x1), math.abs(y2-y1);
					if ( dx <= (cobj.width + obj.width)/2 ) and ( dy <= (cobj.height + obj.height)/2 ) then
						collided = true;
						if ( dx < dy ) then
							if obj.y >= cobj.y then
								t = {'landed', obj, 'bottom', cobj, 'top'}
							else
								t = {'landed', cobj, 'bottom', obj, 'top'}
							end
						else
							if obj.x >= cobj.x then
								t = {'walled', obj, 'left', cobj, 'right'}
							else
								t = {'walled', cobj, 'left', obj, 'right'}
							end
						end
						if obj.collider ~= cobj then
							print(GetTime(), ": collision!", unpack(t));
							if type(self.callback) == "function" then
								local ret, who = self.callback(unpack(t));
								print("Callback: ", ret,who);
								if ret == "BLOCK" then
									if t[1] == 'walled' then
										who.walled = cobj;
										if who == cobj then who.walled = obj; end
										who.landed = nil;
										if x <= cobj.x then x = cobj.x - obj.width; who.limitH = 1; else x = cobj.x + cobj.width who.limitH = -1;end
									else
										who.walled = nil;
										if who == cobj then who.landed = obj; end
										who.landed = cobj; who.grounded = true;
										who.forceH = 0;
										if y <= cobj.y then y = cobj.y - obj.height; who.limitV = 1; else y = cobj.y + cobj.height; who.limitV = -1; end

									end
								end
							end
						end
						obj.collider = cobj;
						cobj.collider = obj;
					end
				end
			end
			if not collided then
				obj.collider = nil;
				obj.walled = nil;
				obj.landed = nil;
				obj.limitV = 0;
				obj.limitH = 0;
			end
			obj.x,obj.y = x,y;
		end
	end
end

local function Gnomario_Physics_Put(self, obj, x, y)
	obj.x = x;
	obj.y = y;
	obj.limitH = 0;
	obj.limitV = 0;
	Apply_Physics(obj);
	table.insert(self.objects, obj);
end

function Gnomario_Physics_Chamber(width,height,callback)
	local chamber = {};
	chamber.width = width;
	chamber.height = height;
	chamber.objects = {};
	chamber.callback = callback;
	chamber.Step = Gnomario_Physics_Step;
	chamber.Add = Gnomario_Physics_Add;
	chamber.Put = Gnomario_Physics_Put;
	return chamber;
end
